#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include "bitar.h"

/***********************************************************************
 *                                                                     *
 *                              BIT ARRAY                              *
 *                                                                     *
 ***********************************************************************/

struct bitar
{
  int len, blen;
  uint64_t *data;
};

bitar_t *bitar_new(int len)
{
  bitar_t *ar;
  ar= malloc(sizeof(*ar));
  ar->blen= len;
  ar->len= (len+63)/64;
  ar->data= calloc(sizeof(ar->data[0]), ar->len);
  return ar;
}

void     bitar_free(bitar_t *ar)
{
  free(ar->data);
  free(ar);
}

int      bitar_get(bitar_t *ar, int pos)
{
  uint64_t p;
  if (pos<0 || pos>=ar->blen) return 0;
  p= pos;
  return (ar->data[p/64] >> (p%64)) & 1;
}

void     bitar_set(bitar_t *ar, int pos)
{
  uint64_t p;
  if (pos<0 || pos>=ar->blen) return ;
  p= pos;
  ar->data[p/64] |= 1ull << (p%64); 
}

void     bitar_clr(bitar_t *ar, int pos)
{
  uint64_t p;
  if (pos<0 || pos>=ar->blen) return ;
  p= pos;
  ar->data[p/64] &= ~(1ull << (p%64)); 
}

void     bitar_setv(bitar_t *ar, int pos, int value)
{
  if (value) bitar_set(ar,pos);
        else bitar_clr(ar,pos);
}


bitar_t *bitar_dup(bitar_t *ar)
{
  bitar_t *cp;
  cp= malloc(sizeof(*cp));
  cp->len= ar->len;
  cp->blen= ar->blen;
  cp->data= malloc(sizeof(cp->data[0])*cp->len);
  memcpy(cp->data, ar->data, sizeof(cp->data[0])*cp->len);
  return cp;
}

void     bitar_clr_all(bitar_t *ar)
{
  memset(ar->data, 0, sizeof(ar->data[0])*ar->len);
}

static int count_bits64(uint64_t A)
{
  uint32_t i,j;
  static const uint64_t B[]=
  {
    0x5555555555555555ll,
    0x3333333333333333ll,
    0x0f0f0f0f0f0f0f0fll,
    0x00ff00ff00ff00ffll,
    0x0000ffff0000ffffll,
    0x00000000ffffffffll
  };
  
  for(j=0,i=1;i<64;i<<=1,j++) A= (A&B[j]) + ((A>>i)&B[j]);
  return A;
}

int      bitar_count1(bitar_t *ar)
{
  int total,i;
  total= 0;
  for(i=0;i<ar->len;i++)
    total+= count_bits64(ar->data[i]);
  return total;
}

int      bitar_count0(bitar_t *ar)
{
  return ar->blen - bitar_count1(ar);
}
