#include <stdio.h> 
#include <stdint.h> 
#include <string.h> 
#include <stdlib.h> 
#include <sys/types.h> 
#include <errno.h> 
#include "intset.h"

typedef struct
{
  union { struct { void* child[2];           } inte;  
          struct { uint32_t key;             } leaf; } d;
  unsigned internal:1, bit:6;
} intset_node_t;

#define Q_KEY(X)    ((X)->d.leaf.key)
#define Q_CHLD(X,Y) ((X)->d.inte.child[Y])


static inline int intset_dir(uint32_t key,intset_node_t *node)
{
  return (key >> node->bit)&1;
}

int intset_find(intset_t *map,uint32_t key)
{
  intset_node_t *node;

  node= map->root;
  if (!node) return 1;

  while(node->internal) node= Q_CHLD(node, intset_dir(key,node));

  return Q_KEY(node)!=key;
}



static inline intset_node_t* intset_leaf_node(uint32_t key)
{
  intset_node_t *N;
  N= malloc(sizeof(*N));
  N->internal= 0;
  Q_KEY(N)= key;
  return N;
}

static inline intset_node_t* intset_internal_node(uint32_t bit)
{
  intset_node_t *N;
  N= malloc(sizeof(*N));
  N->internal= 1;
  N->bit= bit;
  return N;
}

static uint8_t intset_crit_bit(uint32_t X, uint32_t Y)
{
  uint32_t i,j;
  static const uint32_t B[]=
  {
    0x55555555,
    0x33333333,
    0x0f0f0f0f,
    0x00ff00ff,
    0x0000ffff,
  };
  uint32_t A;
  A= X ^ Y;
  
  for(i=1;i<32;i<<=1) A |= (A<<i);
  A = ~A;
  for(j=0,i=1;i<32;i<<=1,j++) A= (A&B[j]) + ((A>>i)&B[j]);
  return A;
}

int intset_insert(intset_t* map, uint32_t key)
{
  intset_node_t *node_elt;

  if (!map->root)
  {
     map->root= intset_leaf_node(key);
     return 0;
  }

  node_elt= map->root;
  while(node_elt->internal)
    node_elt= Q_CHLD(node_elt, intset_dir(key,node_elt));

  
  uint32_t elt;
  int dir_elt;
  uint8_t bit;
  intset_node_t *node_key;

  elt= Q_KEY(node_elt);

  if (elt==key) return 1;

  bit= intset_crit_bit(elt,key);

  dir_elt= (elt>>bit)&1;

  node_key= intset_internal_node(bit);
  Q_CHLD(node_key,1-dir_elt)= intset_leaf_node(key);

  intset_node_t *parent;
  intset_node_t *node;

  parent= 0;
  node= map->root;

  while(node->internal)
  {
     if (node->bit>bit) break;
     parent= node;
     node= Q_CHLD(node, intset_dir(key,node));
  }

  Q_CHLD(node_key,dir_elt)= node;
  if (parent) { if (Q_CHLD(parent,0)==node)  Q_CHLD(parent,0)= node_key;
                                       else  Q_CHLD(parent,1)= node_key; }
  else map->root= node_key;
  return 0;
}

int intset_remove(intset_t* map, uint32_t key)
{
  if (!map->root) return 1;
   
  intset_node_t *node;
  intset_node_t *sibling;
  intset_node_t *parent;
  intset_node_t *gparent;

  node= map->root;
  gparent= 0;
  sibling= 0;
  parent= 0;

  while(node->internal)
  {
    int dir;
    dir= intset_dir(key, node);
    gparent= parent;
    parent= node;
    sibling= Q_CHLD(node, 1-dir);
    node= Q_CHLD(node, dir);
  }

  if (Q_KEY(node)!=key) return 1;
  free(node);

  if (!gparent)
  {
     map->root= sibling;
     if (parent) free(parent);
     return 0;
  }

  if (Q_CHLD(gparent,0)==parent) Q_CHLD(gparent,0)= sibling;
                            else Q_CHLD(gparent,1)= sibling;
  free(parent);

  return 0;
}

int intset_destroy(intset_t *map,uint32_t *Rkey)
{
  intset_node_t *node;
  intset_node_t *parent;
  intset_node_t *gparent;
  intset_node_t *sibling;

  if (!map->root) return 1;

  parent= 0;
  gparent= 0;
  sibling= 0;
  node= map->root;
  while(node->internal)
  {
    gparent= parent;
    parent= node;
    sibling= Q_CHLD(node, 1);
    node= Q_CHLD(node,0);
  }
  if (Rkey) *Rkey= Q_KEY(node);
  free(node);
  if (!gparent) 
  {
    map->root= sibling;
    if (parent) free(parent);
    return 0;
  }
  if (Q_CHLD(gparent,0)==parent)
    Q_CHLD(gparent,0)= sibling;
  else
    Q_CHLD(gparent,1)= sibling;

  free(parent);

  return 0;
}

intset_t  *intset_new()
{
  intset_t *s;
  s= malloc(sizeof(*s));
  s->root= 0;
  return s;
}
