
/***********************************************************************
 *                                                                     *
 *                            Manual Search                            *
 *                                                                     *
 ***********************************************************************
 * This file implements some 'manual search' operators. These are      *
 * candidates for automatic search operators and stowed here just to   *
 * make space in main.c for my new generator stuff.                    *
 ***********************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include "gfx.h"
#include "network.h"
#include "draw-net.h"
#include "main.h"
#include "manser.h"


typedef struct { int x,y,z; } triplet_t;
#define DENSMAX 500
static triplet_t dens[DENSMAX];
static int ndens;


static char *txtfiles[]=
{
  "Text Files (.txt)",
  NULL
};

#if 0
static int choose_prefix_existing_pair
  (network_t *NW,int *Ra,int *Rb,int nr,unsigned char *remaining,
                                        unsigned char *visited)
{
  // First of all, we want to eliminate single-gate levels. Let's see
  // if we can find any.
  int i;
  compex_t *CE;
  for(i=0;i<NW->ngates;i++)
  {
    CE= NW->gates[i];
    if (CE->level==0) continue;
    if (NW->gates[i-1]->level!=CE->level &&
        (i==NW->ngates-1 || NW->gates[i+1]->level!=CE->level)
        && !visited[CE->A] && !visited[CE->B])
    {
          *Ra= CE->A;
          *Rb= CE->B;
          return 1;
    }
  }
  return 0;
}
#endif

static void choose_prefix(network_t *NW,int *Ra, int *Rb)
{
  unsigned char remaining[200];
  unsigned char visited[200];
  int i,j,nr;
  for(i=0;i<NW->ninputs;i++) visited[i]= 0; 
  for(i=0;i<NW->ngates;i++)
    if (NW->gates[i]->level!=0) break;
    else { visited[NW->gates[i]->A]= visited[NW->gates[i]->B]= 1; }
  nr= 0;
  for(i=0;i<NW->ninputs;i++)
    if (!visited[i]) remaining[nr++]= i;

  if (nr<2)
  {
    i= rand() % NW->ninputs;
    for(j=i;j==i;) j= rand()%NW->ninputs;
    *Ra= i; *Rb= j;
    return ;
  }
#if 0
  if (choose_prefix_existing_pair(NW,Ra,Rb,nr,remaining,visited)) return ;
#endif

  i= rand() % nr;
  *Ra= remaining[i];
  remaining[i]= remaining[nr-1];
  nr--;
  if (nr==1) { *Rb= remaining[0]; return ; }
  j= rand() % nr;
  *Rb= remaining[j];
}

static void get_level_avail(gatelist_t *GL,int level,int *Rna,int *avail)
{
  int i;
  char visited[64];
  int N;
  int na;
  gate_t *g;
  N= GL->ninputs;
  for(i=0;i<N;i++) visited[i]= 0;
  for(i=0;i<GL->ngates;i++)
  {
    g= GL->gates+i;
    if (g->level==level) visited[g->A]= visited[g->B]= 1;
    else if (g->level>level) break;
  }
  na= 0; 
  for(i=0;i<N;i++)
    if (!visited[i]) avail[na++]= i;
  *Rna= na;
}

static network_t *make_network_with_insertion
  (gatelist_t *GL,int level,int a,int b)
{
  int i;
  network_t *WW;
  gate_t *g;
  WW= network_new(GL->ninputs);
  for(i=0;i<GL->ngates;i++)
  {
    g= GL->gates+i;
    if (g->level==level)
    {
      network_append_gate(WW, a,b);
      level= -1;
    }
    network_append_gate(WW, g->A, g->B);
  }
  return WW;
}


static void remove_something_fwd()
{
  gatelist_t *GL;
  int i,j;
  network_t *WW;
  if (!NW) return ;
  network_to_gate_list(NW,&GL);
  for(i=0;i<GL->ngates;i++)
  {
    if (GL->gates[i].level==0) continue;
    WW= network_new(GL->ninputs);
    for(j=0;j<GL->ngates;j++)
     if (j!=i) network_append_gate(WW,GL->gates[j].A,GL->gates[j].B);
    if (network_sorts(WW))
    {
      for(j=0;j<GL->ngates;j++)
        if (j==i) NW->gates[j]->color=0xff0000;
             else NW->gates[j]->color=0x000000;
      show_network(0);
      network_free(WW);
      goto done;
    }
    network_free(WW);
  }
done:
  gatelist_free(GL);
}
static void remove_something()
{
  gatelist_t *GL;
  int i,j;
  network_t *WW;
  if (!NW) return ;
  network_to_gate_list(NW,&GL);
  for(i=GL->ngates-1;i>=0;i--)
  {
    WW= network_new(GL->ninputs);
    for(j=0;j<GL->ngates;j++)
     if (j!=i) network_append_gate(WW,GL->gates[j].A,GL->gates[j].B);
    if (network_sorts(WW))
    {
      for(j=0;j<GL->ngates;j++)
        if (j==i) NW->gates[j]->color=0xff0000;
             else NW->gates[j]->color=0x000000;
      show_network(0);
      network_free(WW);
      goto done;
    }
    network_free(WW);
  }
done:
  gatelist_free(GL);
}

static int densify()
{
  gatelist_t *GL;
  int i,j,k,A,B,L;
  network_t *WW;
  int avail[50]; int navail;

  if (!NW) return 0;
  network_to_gate_list(NW,&GL);

  ndens= 0;
  for(i=1;i<=GL->gates[GL->ngates-1].level;i++)
  {
    get_level_avail(GL, i, &navail, avail);
    for(j=0;j<navail;j++)
     for(k=j+1;k<navail;k++)
      { if (ndens==DENSMAX) goto select_again;
        dens[ndens].x= avail[j];
        dens[ndens].y= avail[k];
        dens[ndens].z= i; 
        ndens++; }
  }

select_again:
  if (ndens==0) goto fail;
  k= rand()%ndens;
  A= dens[k].x;
  B= dens[k].y;
  L= dens[k].z;
  dens[k]= dens[ndens-1];
  ndens--;

  WW= make_network_with_insertion(GL, L, A,B);
  if (network_sorts(WW))
  {
    gatelist_free(GL);
    replace_network(WW);
    return 1;
  }
  network_free(WW);
  goto select_again;

fail:
  return 0;
}

static int normalize_gatelist_reverse(gatelist_t *GL)
{
  int change, nextlevel, curlevel,i,j,N;
  char L[200];
  gate_t t, *g;

  change= 0;
  N= GL->ninputs;

  for(nextlevel=GL->gates[GL->ngates-1].level;nextlevel>0;nextlevel--)
  {
    for(i=0;i<N;i++) L[i]= 0;
    for(i=GL->ngates-1;i>=0;i--)
    {
       g= GL->gates+i;
       if (g->level>nextlevel) continue;
       if (g->level==nextlevel)
          L[g->A]= L[g->B]= 1;
       else break;
    }

    curlevel= nextlevel-1;
    for(j=i;j>=0;j--)
    {
      g= GL->gates+j;
      if (g->level!=curlevel) break;
      if (!L[g->A] && !L[g->B])
      {
        g->level= nextlevel;
        L[g->A]= L[g->B]= 1;
        t= GL->gates[i];
        GL->gates[i]= GL->gates[j];
        GL->gates[j]= t;
        i--;
        change= 1;
      }
    }
  }
  return change;
}

static void print_gatelist(gatelist_t *GL)
{
  int i;
  for(i=0;i<GL->ngates;i++) printf("%d.%d[%d] ",
        GL->gates[i].A,GL->gates[i].B,GL->gates[i].level);
  printf("\n");
  fflush(stdout);
}

static int normalize_gatelist(gatelist_t *GL)
{
  int change, prevlevel, curlevel,i,j,N;
  char L[200];
  gate_t t, *g;

  change= 0;
  N= GL->ninputs;

  for(prevlevel=0;prevlevel<GL->gates[GL->ngates-1].level;prevlevel++)
  {
    for(i=0;i<N;i++) L[i]= 0;
    for(i=0;i<GL->ngates;i++)
    {
       g= GL->gates+i;
       if (g->level<prevlevel) continue;
       if (g->level==prevlevel)
          L[g->A]= L[g->B]= 1;
       else break;
    }

    curlevel= prevlevel+1;
    for(j=i;j<GL->ngates;j++)
    {
       g= GL->gates+j;
       if (g->level!=curlevel) break;
       if (!L[g->A] && !L[g->B])
       {
         L[g->A]= L[g->B]= 1;
         t= GL->gates[i];
         GL->gates[i]= GL->gates[j];
         GL->gates[j]= t;
         GL->gates[i].level= prevlevel;
         i++;
         change= 1;
       }
    }
  }
  return change;
}

static void print_gatelist_levels(gatelist_t *GL)
{
  int i;
  for(i=0;i<GL->ngates;i++) printf("%2d ", GL->gates[i].level);
  printf("\n");
}

static void do_normalize(int right)
{
  gatelist_t *GL;
  if (!NW) return;

  network_to_gate_list(NW,&GL);

  if (right) while(normalize_gatelist_reverse(GL));
       else  while(normalize_gatelist(GL));

  replace_network(network_from_gate_list(GL,right));
  gatelist_free(GL);
}

static void choose_prefix_for_last()
{
  int i,j,k;
  int N;
  int lastA, lastB, ll;
  int nu, u[500];
  network_t *WW;
  compex_t *CE;
  int maxnu;
  network_t *best;

  maxnu= 1;
  best= NULL;

  N= NW->ninputs;
  lastA= NW->gates[NW->ngates-1]->A;
  lastB= NW->gates[NW->ngates-1]->B;

  for(i=0;i<N;i++)
   for(j=i+1;j<N;j++)
   {
     WW= network_new(N);
     network_append_gate(WW, i,j);
     for(k=0;k<NW->ngates;k++)
       network_append_gate(WW, NW->gates[k]->A, NW->gates[k]->B);
     network_all_useless_gates(WW,&nu,u);
     if (nu==0)
     {
       network_free(WW);
       continue;
     }
     network_levels(WW);
     ll= WW->gates[WW->ngates-1]->level;
     for(k=0;k<nu;k++)
     {
       CE= WW->gates[u[k]];
       if (CE->level==ll && CE->A==lastA && CE->B==lastB)
       {
          replace_network(WW);
          if (best) network_free(best);
          return ;
       }
     }
     if (nu>maxnu) { maxnu= nu; if (best) network_free(best);
                                best= WW; }                     
     else network_free(WW);
   }

   if (best) replace_network(best);
}

static network_t *reverse_network(network_t *ZZ)
{
  network_t *WW;
  int i;
  gatelist_t *GL;

  network_to_gate_list(ZZ,&GL);
  WW=network_new(GL->ninputs);
  for(i=GL->ngates;i>=0;i--)
    network_append_gate(WW,GL->gates[i].A,GL->gates[i].B);
  gatelist_free(GL);
  if (!network_sorts(WW))
  {
    network_free(WW);
    return NULL;
  }
  return WW;
}

static int flip_and_show_useless()
{
  network_t *WW;
  int K;
  if (!NW) return 0;

  WW= reverse_network(NW);
  if (!WW) return 0;

  K= network_find_useless_gate(WW);
  if (K==-1) { network_free(WW); return 0; }
  WW->gates[K]->color= 0xff0000;
  replace_network(WW);
  return 1;
}

static void reproduce()
{
  network_t *WW;
  gatelist_t *GL;
  int i;
  if (!NW) return ;
  network_to_gate_list(NW,&GL);
  WW= network_new(GL->ninputs);
  for(i=0;i<GL->ngates;i++)
    network_append_gate(WW,GL->gates[i].A,GL->gates[i].B);
  for(i=GL->ngates;i>=0;i--)
    network_append_gate(WW,GL->gates[i].A,GL->gates[i].B);
  gatelist_free(GL);
  replace_network(WW);
}

static void remove_redgates()
{
  int j;
  network_t *WW;
  if (!NW) return ;
  WW= network_new(NW->ninputs);
  for(j=0;j<NW->ngates;j++)
    if (NW->gates[j]->color!=0xff0000)
      network_append_gate(WW,NW->gates[j]->A,NW->gates[j]->B);
  replace_network(WW);
}

static void repromax()
{
  reproduce();
  while(flip_and_show_useless())
   remove_redgates();
}

static void swap_simple()
{
  int S;
  gatelist_t *GL;
  int i,j,k,R;
  exp_t *min,*max,*v;
  network_t *WW;
  int top,bot,mid;
  int *pairs;
  int npairs;
  int G1,G2;
  int tries;

  tries= 0;

  if (!NW) return ;

  pairs= malloc(sizeof(int)*NW->ngates*NW->ngates*2);
  npairs= 0;
  for(i=0;i<NW->ngates;i++)
   for(j=i+1;j<NW->ngates;j++)
      { pairs[2*npairs]= i;
        pairs[2*npairs+1]= j;
        npairs++; }

  for(i=0;i<npairs;i++)
  {
    compex_t *C1, *C2;
    C1= NW->gates[pairs[2*i]];
    C2= NW->gates[pairs[2*i+1]];

    if (//C1->level==C2->level ||
        (C1->A==C2->A && C1->B==C2->B) )
    {
      pairs[2*i]= pairs[2*(npairs-1)];
      pairs[2*i+1]= pairs[2*(npairs-1)+1];
      npairs--;
      i--;
    }
  }

again:
  for(i=0;i<tries;i++) printf(" ");
  printf("%d swappable pairs\n",npairs);

  if (npairs==0) { free(pairs); return ; }

  S= rand()%npairs;

  G1= pairs[2*S];
  G2= pairs[2*S+1];
  
  network_to_gate_list(NW,&GL);

      gate_t t;
  t= GL->gates[G1];
  GL->gates[G1]= GL->gates[G2];
  GL->gates[G2]= t;

  WW= network_from_gate_list(GL,0);
  gatelist_free(GL);
      
  R= network_sorts(WW);
  if (R) { free(pairs); replace_network(WW);  return ; }
  network_free(WW);  

  pairs[2*S]= pairs[2*(npairs-1)];
  pairs[2*S+1]= pairs[2*(npairs-1)+1];
  npairs--;
  tries++;
  goto again;
}

static int replace_gate(gatelist_t *src, int ex, unsigned int mask)
{
  gatelist_t CP;
  gate_t gates[500];
  network_t *WW;
  int i,j,k,z,N;

  CP.gates= gates;
  CP.ngates= src->ngates;
  N= CP.ninputs= src->ninputs;

  for(i=0;i<N-1;i++)
  {
   // if (mask&(1<<i)) continue;
   for(j=i+1;j<N;j++)
   {
      // if (mask&(1<<j)) continue;
      gates[0].A= i;
      gates[0].B= j;
      z= 1;
      for(k=0;k<src->ngates;k++)
        if (k!=ex) gates[z++]= src->gates[k];
      WW= network_from_gate_list(&CP, 0);
      if (network_sorts(WW)) { replace_network(WW); return 1; }
      network_free(WW);
   }
  }

  return 0;
}

static void remove_single()
{
  gatelist_t *GL;
  int i;
  int nex;
  int exlist[500];
  unsigned int mask0;
  int pdif,ndif;

  if (!NW) return ;
  network_to_gate_list(NW,&GL);

  nex= 0;
  for(i=GL->ngates-1;i>0;i--)
  {
    ndif= (i==GL->ngates-1) || (GL->gates[i].level!=GL->gates[i+1].level);
    pdif= (GL->gates[i].level!=GL->gates[i-1].level);
    if (ndif && pdif)
      exlist[nex++]= i;
  }
  mask0= 0;
  for(i=0;i<GL->ngates;i++)
   if (GL->gates[i].level==0)
     mask0 |= (1<<GL->gates[i].A) | (1<<GL->gates[i].B);
  for(i=0;i<nex;i++)
    if (replace_gate(GL, exlist[i], mask0))
      break;
  gatelist_free(GL);
}

static void find_shuffle()
{
  int K;
  int i;
  if (!NW) return ;
  for(i=0;i<NW->ngates;i++) NW->gates[i]->color= 0;
  K= network_find_shuffle_gate(NW);
  if (K!=-1) 
    NW->gates[K]->color= 0xff;
  show_network(0);
}

static int replace_int(int *tgt, int x, int y)
{
  if (*tgt!=x) return 0;
  *tgt= y;
  return 1;
}

static void order_ints(int *l, int *h)
{
  int t;
  if (*l<*h) return ;
  t = *l;
  *l= *h;
  *h= t;
}

static void shuffle()
{
  int K,A,B;
  int i;
  gatelist_t *GL;
  network_t *WW;
  if (!NW) return ;
  for(i=0;i<NW->ngates;i++) if (NW->gates[i]->color==0xff) break;
  if (i==NW->ngates) return ;
  K= i;
  network_to_gate_list(NW,&GL);
  A= GL->gates[K].A;
  B= GL->gates[K].B;
  for(i=0;i<K;i++)
  {
    if (!replace_int(&GL->gates[i].A, A, B)) 
      replace_int(&GL->gates[i].A, B, A);
    if (!replace_int(&GL->gates[i].B, A, B))
      replace_int(&GL->gates[i].B, B, A);
    order_ints(&GL->gates[i].A, &GL->gates[i].B);
  }
  WW= network_from_gate_list(GL, 0);
  gatelist_free(GL);
  replace_network(WW);
}

static void double_network()
{
  int N,N2,i;
  network_t *WW;
  int A,B;
  if (!NW) return ;
  N= NW->ninputs;
  N2= N*2;
  WW= network_new(N2);
  for(i=0;i+1<N2;i+=2) network_append_gate(WW,i,i+1);
#if 0
  for(i=0;i<NW->ngates;i++)
  {
    A= NW->gates[i]->A;
    B= NW->gates[i]->B;
    network_append_gate(WW,A,B);
    network_append_gate(WW,A+N,B+N);
  }
  for(i=0;i<NW->ngates;i++)
  {
    A= NW->gates[i]->A;
    B= NW->gates[i]->B;
    network_append_gate(WW,A+N/2,B+N/2);
  }
  for(i=0;i<NW->ngates;i++)
  {
    A= NW->gates[i]->A;
    B= NW->gates[i]->B;
    network_append_gate(WW,A,B);
    network_append_gate(WW,A+N,B+N);
  }
  for(i=0;i<NW->ngates;i++)
  {
    A= NW->gates[i]->A;
    B= NW->gates[i]->B;
    network_append_gate(WW,A+N/2,B+N/2);
  }
#endif
#if 1
  for(i=0;i<NW->ngates;i++)
  {
    A= NW->gates[i]->A;
    B= NW->gates[i]->B;
    network_append_gate(WW, 2*A, 2*B);
    network_append_gate(WW, 2*A+1, 2*B+1);
  }
#endif
  for(i=1;i+1<N2;i+=2) network_append_gate(WW,i,i+1);
  replace_network(WW);
}

static void halve_network()
{
  int N,N2,i;
  network_t *WW;
  if (!NW) return ;
  N= NW->ninputs;
  N2= N/2;
  WW= network_new(N2);
  for(i=0;i<NW->ngates;i++)
  {
    int A,B;
    A= NW->gates[i]->A;
    B= NW->gates[i]->B;
    if (A<N2 && B<N2)
      network_append_gate(WW, A, B);
  }
  replace_network(WW);
}

int bottom_event(gfx_event_t event)
{
  char *fn;
  int i,j,u[400],nu;
  network_t *WW;

  switch(event.type)
  {
  case Wuseless:
     if (!NW) return 1;
     network_all_useless_gates(NW,&nu,u);
     if (nu==0) return 1;
     for(i=0;i<nu;i++)
        NW->gates[u[i]]->color= 0xff0000;
     show_network(0);
     return 1;
  case Wremove:
     remove_redgates();
     return 1;
  case Wrandomprefix:
     if (!NW) return 1;
     WW= network_new(NW->ninputs);
     choose_prefix(NW, &i,&j);
     network_append_gate(WW,i,j);
     for(j=0;j<NW->ngates;j++)
       network_append_gate(WW,NW->gates[j]->A,NW->gates[j]->B);
     replace_network(WW);
     return 1;
  case Wnormalize: do_normalize(0); return 1;
     case Wrnormalize: do_normalize(1); return 1;
     case Wtryremove: remove_something(); return 1;
     case Wtryfwd: remove_something_fwd(); return 1;
   
     case Wsave:
       if (!NW) return 1;
       fn= gfx_select_file(1,NULL,txtfiles);
       if (fn) network_to_file(NW, fn);
       return 1;
     case Wdensify: while(densify()); return 1;
     case Wsorts: if (NW && network_sorts(NW)) gfx_set_text(Wstats,"Sorts");
				 else gfx_set_text(Wstats,"Doesn't sort");
          return 1;
     case Wremovefinal: 
          choose_prefix_for_last(); return 1;
     case Wreproduce: reproduce(); return 1;
     case Wflipuseless: flip_and_show_useless(); return 1;
     case Wswap: swap_simple(); return 1;
     case Wrmsingle: remove_single(); return 1;
     case Wfindshuf: find_shuffle(); return 1;
     case Wshuf: shuffle(); return 1;
     case Wrepromax: repromax(); return 1;
     case Wprepwall:
       WW= network_new(NW->ninputs);
       for(i=0;2*i+1<NW->ninputs;i++)
         network_append_gate(WW, 2*i,2*i+1);
       for(i=0;i<NW->ngates;i++)
         network_append_gate(WW, NW->gates[i]->A, NW->gates[i]->B);
       replace_network(WW);
       return 1;
     case Wwall:
       WW= network_new(NW->ninputs);
       for(i=0;i<NW->ngates;i++)
         network_append_gate(WW, NW->gates[i]->A, NW->gates[i]->B);
       for(i=0;2*i+1<NW->ninputs;i++)
         network_append_gate(WW, 2*i,2*i+1);
       replace_network(WW);
       return 1;
     case Wdouble: double_network(); return 1;
     case Whalf: halve_network(); return 1;
    }
  return 0;
}
