#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include "util.h"

int bit_count(uint32_t A)
{
  int i,j;
  unsigned int mask[5]=
  {
    0x55555555,
    0x33333333,
    0x0f0f0f0f,
    0x00ff00ff,
    0x0000ffff
  };
  for(i=0,j=1;i<5;i++,j<<=1)
    A= (A&mask[i]) + ((A>>j)&mask[i]);
  return A;
}

int highest_bit(uint32_t A)
{
  uint32_t i;
  for(i=1;i<32;i<<=1) A |= (A>>i);
  return (bit_count(A)+32)%33;    
}

int lowest_bit(uint32_t A)
{
  uint32_t i;
  for(i=1;i<32;i<<=1) A |= (A<<i);
  return bit_count(~A);
}

int imax2(int a,int b)
{
  return a>b?a:b;
}
int imin2(int a,int b)
{
  return a<b?a:b;
}


static void check_down32(int x, int N, uint32_t *A)
{
  int L,R,G;
  uint32_t t;
  while(x<N/2) 
  {
    L= 2*x+1;
    R= 2*x+2;
    G=L;
    if (R<N && A[R]>A[L]) G= R;
    if (A[x]>A[G]) return ;
    t= A[x]; A[x]= A[G]; A[G]= t;
    x= G;
  }
}

void heapsort_uint32(int N, uint32_t *A)
{
  int i;
  uint32_t t;

  for(i=N/2-1;i>=0;i--) check_down32(i, N, A);

  while(N)
  {
    t= A[N-1]; A[N-1]= A[0]; A[0]= t;
    N--;
    check_down32(0,N,A);
  }
}

static int unify_uint32(int N,uint32_t *A)
{
  int i,j;
  if (N<2) return N;
  j= 0;
  for(i=1;i<N;i++)
   if (A[i]!=A[j])  A[++j]= A[i]; 
  return j+1;
}


void unify_intlist(intlist_t *L)
{
  heapsort_uint32(L->len,L->data);
  L->len= unify_uint32(L->len,L->data);
}


static void check_down16(int x, int N, uint16_t *A)
{
  int L,R,G;
  uint16_t t;
  while(x<N/2) 
  {
    L= 2*x+1;
    R= 2*x+2;
    G=L;
    if (R<N && A[R]>A[L]) G= R;
    if (A[x]>A[G]) return ;
    t= A[x]; A[x]= A[G]; A[G]= t;
    x= G;
  }
}

void heapsort_uint16(int N, uint16_t *A)
{
  int i;
  uint16_t t;

  for(i=N/2-1;i>=0;i--) check_down16(i, N, A);

  while(N)
  {
    t= A[N-1]; A[N-1]= A[0]; A[0]= t;
    N--;
    check_down16(0,N,A);
  }
}

static int unify_uint16(int N,uint16_t *A)
{
  int i,j;
  if (N<2) return N;
  j= 0;
  for(i=1;i<N;i++)
   if (A[i]!=A[j])  A[++j]= A[i]; 
  return j+1;
}


void unify_sholist(sholist_t *L)
{
  heapsort_uint16(L->len,L->data);
  L->len= unify_uint16(L->len,L->data);
}

static void read_bytes(FILE *f, void *Z, int len)
{
  fread(Z, 1,len,f);
}
static void write_bytes(FILE *f, void *Z, int len)
{
  fwrite(Z, 1,len,f);
}

static void *read_bytes_a(FILE *f, int N)
{
  void *K;
  K= malloc(N);
  read_bytes(f, K, N);
  return K;
}

void intlist_write(FILE *f,int N, uint32_t *data)
{
  write_bytes(f, data, N*sizeof(data[0]));
}

uint32_t *intlist_read(FILE *f,int N)
{
  return read_bytes_a(f, N*sizeof(uint32_t));
}

void sholist_write(FILE *f,int N, uint16_t *data)
{
  write_bytes(f, data, N*sizeof(data[0]));
}

uint16_t* sholist_read(FILE *f,int N)
{
  return read_bytes_a(f, N*sizeof(uint16_t));
}

void read_int(FILE *f,uint32_t *R) { read_bytes(f, R, sizeof(R[0])); }
void write_int(FILE *f,uint32_t R) { write_bytes(f, &R, sizeof(R)); }


sholist_t *sholist_new(int len)
{
  sholist_t *L;
  L= malloc(sizeof(*L));
  L->len= len;
  L->data= malloc(sizeof(L->data[0])*len);
  return L;
}
void sholist_free(sholist_t *L) { free(L->data); free(L); }

